import json
from pathlib import Path
import yaml
import zmq

from tomato import tomato
from . import utils

PORT = 12345
CTXT = zmq.Context()
timeout = 1000
kwargs = dict(port=PORT, context=CTXT, timeout=timeout)


def test_reload_noop(datadir, start_tomato_daemon, stop_tomato_daemon):
    assert utils.wait_until_tomato_running(port=PORT, timeout=timeout)
    ret = tomato.reload(**kwargs, appdir=Path())
    assert ret.success
    assert len(ret.data.drvs) == 1
    assert len(ret.data.devs) == 1
    assert len(ret.data.pips) == 1
    assert len(ret.data.cmps) == 1


def test_reload_settings(datadir, start_tomato_daemon, stop_tomato_daemon):
    assert utils.wait_until_tomato_running(port=PORT, timeout=timeout)

    with open("settings.toml", "a") as inf:
        inf.write("example_counter.testparb = 1")
    ret = tomato.reload(**kwargs, appdir=Path())
    print(f"{ret=}")
    assert ret.success
    assert len(ret.data.drvs) == 1
    assert len(ret.data.devs) == 1
    assert len(ret.data.pips) == 1
    assert len(ret.data.cmps) == 1
    assert ret.data.drvs["example_counter"].settings["testparb"] == 1


def test_reload_cmps_pips(datadir, start_tomato_daemon, stop_tomato_daemon):
    assert utils.wait_until_tomato_running(port=PORT, timeout=timeout)

    with open("devices_counter.json", "r") as inf:
        jsdata = json.load(inf)
    with open("devices.yml", "w") as ouf:
        yaml.dump(jsdata, ouf)

    ret = tomato.reload(**kwargs, appdir=Path())
    print(f"{ret=}")
    assert ret.success
    assert len(ret.data.drvs) == 1
    assert len(ret.data.devs) == 1
    assert len(ret.data.pips) == 4
    assert len(ret.data.cmps) == 4


def test_reload_devs(datadir, start_tomato_daemon, stop_tomato_daemon):
    assert utils.wait_until_tomato_running(port=PORT, timeout=timeout)

    with open("devices_multidev.json", "r") as inf:
        jsdata = json.load(inf)
    with open("devices.yml", "w") as ouf:
        yaml.dump(jsdata, ouf)

    ret = tomato.reload(**kwargs, appdir=Path())
    print(f"{ret=}")
    assert ret.success
    assert len(ret.data.drvs) == 1
    assert len(ret.data.devs) == 2
    assert len(ret.data.pips) == 2
    assert len(ret.data.cmps) == 3


def test_reload_drvs(datadir, start_tomato_daemon, stop_tomato_daemon):
    assert utils.wait_until_tomato_running(port=PORT, timeout=timeout)
    utils.wait_until_tomato_drivers(port=PORT, timeout=3000)

    # Let's add psutil driver / device
    with open("devices_psutil.json", "r") as inf:
        jsdata = json.load(inf)
    with open("devices.yml", "w") as ouf:
        yaml.dump(jsdata, ouf)

    ret = tomato.reload(**kwargs, appdir=Path())
    print(f"{ret=}")
    assert ret.success
    assert len(ret.data.drvs) == 2
    assert len(ret.data.devs) == 2
    assert len(ret.data.pips) == 1
    assert len(ret.data.cmps) == 2
    assert utils.wait_until_tomato_running(port=PORT, timeout=timeout)
    utils.wait_until_tomato_drivers(port=PORT, timeout=3000)

    # Let's remove psutil driver / device and modify channels
    with open("devices_counter.json", "r") as inf:
        jsdata = json.load(inf)
    with open("devices.yml", "w") as ouf:
        yaml.dump(jsdata, ouf)

    ret = tomato.reload(**kwargs, appdir=Path())
    print(f"{ret=}")
    assert ret.success
    assert len(ret.data.drvs) == 1
    assert len(ret.data.devs) == 1
    assert len(ret.data.pips) == 4
    assert len(ret.data.cmps) == 4
    assert utils.wait_until_tomato_running(port=PORT, timeout=timeout)


def test_reload_running(datadir, start_tomato_daemon, stop_tomato_daemon):
    assert utils.wait_until_tomato_running(port=PORT, timeout=timeout)

    utils.run_casenames(["counter_20_5"], [None], ["pip-counter"])
    assert utils.wait_until_ketchup_status(jobid=1, status="r", port=PORT, timeout=5000)

    # Try modifying settings of a driver in use
    with open("settings.toml", "a") as inf:
        inf.write("example_counter.testparb = 1")
    ret = tomato.reload(**kwargs, appdir=Path())
    print(f"{ret=}")
    assert ret.success is False
    assert "reload would modify a driver of a device in a running pipeline" in ret.msg

    # Revert settings.toml back
    with open("settings.toml", "r") as inf:
        lines = inf.readlines()
    with open("settings.toml", "w") as out:
        out.writelines(lines[:-1])
    ret = tomato.reload(**kwargs, appdir=Path())
    print(f"{ret=}")
    assert ret.success

    # Try modifying device driver
    with open("devices_reload_driver.json", "r") as inf:
        jsdata = json.load(inf)
    with open("devices.yml", "w") as ouf:
        yaml.dump(jsdata, ouf)
    ret = tomato.reload(**kwargs, appdir=Path())
    print(f"{ret=}")
    assert ret.success is False
    assert "reload would modify components of a running pipeline" in ret.msg

    # Try removing channel on device
    with open("devices_reload_channel.json", "r") as inf:
        jsdata = json.load(inf)
    with open("devices.yml", "w") as ouf:
        yaml.dump(jsdata, ouf)
    ret = tomato.reload(**kwargs, appdir=Path())
    print(f"{ret=}")
    assert ret.success is False
    assert "reload would modify components of a running pipeline" in ret.msg

    # Try modifying address on device
    with open("devices_reload_address.json", "r") as inf:
        jsdata = json.load(inf)
    with open("devices.yml", "w") as ouf:
        yaml.dump(jsdata, ouf)
    ret = tomato.reload(**kwargs, appdir=Path())
    print(f"{ret=}")
    assert ret.success is False
    assert "reload would modify components of a running pipeline" in ret.msg

    # Try removing pipeline
    with open("devices_reload_pipdel.json", "r") as inf:
        jsdata = json.load(inf)
    with open("devices.yml", "w") as ouf:
        yaml.dump(jsdata, ouf)
    ret = tomato.reload(**kwargs, appdir=Path())
    print(f"{ret=}")
    assert ret.success is False
    assert "reload would delete a running pipeline" in ret.msg

    # Try modifying pipeline
    with open("devices_reload_pipmod.json", "r") as inf:
        jsdata = json.load(inf)
    with open("devices.yml", "w") as ouf:
        yaml.dump(jsdata, ouf)
    ret = tomato.reload(**kwargs, appdir=Path())
    print(f"{ret=}")
    assert ret.success is False
    assert "reload would modify components of a running pipeline" in ret.msg



"""
**tomato.tomato**: command line interface to the tomato daemon
--------------------------------------------------------------
.. codeauthor::
    Peter Kraus

Module of functions to interact with tomato. Includes basic tomato daemon functions:

- :func:`status` to query the status of the tomato daemon
- :func:`start` to start a new tomato daemon
- :func:`stop` to stop a running tomato daemon
- :func:`init` to create a default ``settings.toml`` file
- :func:`reload` to process the ``settings.toml`` and ``devices.yml`` files again

Also includes the following *pipeline* management functions:

- :func:`pipeline_load` to load a sample into a pipeline
- :func:`pipeline_eject` to eject any sample from a pipeline
- :func:`pipeline_ready` to mark a pipeline as ready

"""

import os
import subprocess
import textwrap
import json
from pathlib import Path
from datetime import datetime, timezone
from importlib import metadata

import logging
import psutil
import zmq
import yaml
import toml

from tomato.models import Reply, Pipeline, Device, Driver, Component, Daemon
from tomato.daemon.jobdb import jobdb_setup

logger = logging.getLogger(__name__)
VERSION = metadata.version("tomato")
MAX_RETRIES = 10


def set_loglevel(delta: int):
    loglevel = min(max(30 - (10 * delta), 10), 50)
    logging.basicConfig(level=loglevel)
    logger.debug("loglevel set to '%s'", logging._levelToName[loglevel])


def load_device_file(yamlpath: Path) -> dict:
    logger.debug("loading device file from '%s'", yamlpath)
    try:
        with yamlpath.open("r") as infile:
            jsdata = yaml.safe_load(infile)
    except FileNotFoundError:
        logger.error("device file not found. Running with default devices.")
        devpath = Path(__file__).parent / ".." / "data" / "default_devices.json"
        with devpath.open() as inp:
            jsdata = json.load(inp)
        logger.debug("writing default devices to '%s'", yamlpath)
        with yamlpath.open("w") as outfile:
            yaml.dump(jsdata, outfile)
    return jsdata


def get_pipelines(
    devs: dict[str, Device], pipelines: list
) -> tuple[dict[str, Pipeline], dict[str, Component]]:
    pips = {}
    cmps = {}
    for pip in pipelines:
        if "*" in pip["name"]:
            data = {"name": pip["name"], "devs": {}}
            if len(pip["devices"]) > 1:
                logger.error("more than one component in a wildcard pipeline")
                continue
            for comp in pip["devices"]:
                if comp["device"] not in devs:
                    logger.error("device '%s' not found", comp["device"])
                    break
                dev = devs[comp["device"]]
                for ch in dev.channels:
                    name = pip["name"].replace("*", f"{ch}")
                    h = f"{dev.driver}:({dev.address},{ch})"
                    c = Component(
                        name=h,
                        driver=dev.driver,
                        device=dev.name,
                        address=dev.address,
                        channel=ch,
                        role=comp["role"],
                    )
                    cmps[h] = c
                    p = Pipeline(name=name, components=[h])
                    pips[p.name] = p
        else:
            data = {"name": pip["name"], "components": []}
            for comp in pip["devices"]:
                if comp["device"] not in devs:
                    logger.error("device '%s' not found", comp["device"])
                    break
                dev = devs[comp["device"]]
                if isinstance(comp["channel"], int):
                    logger.warning(
                        "Supplying 'channel' as an int is deprecated "
                        "and will stop working in tomato-2.0."
                    )
                    comp["channel"] = str(comp["channel"])
                if comp["channel"] not in dev.channels:
                    logger.error(
                        "channel %s not found on device '%s'",
                        comp["channel"],
                        comp["device"],
                    )
                    break
                h = f"{dev.driver}:({dev.address},{comp['channel']})"
                c = Component(
                    name=h,
                    driver=dev.driver,
                    device=dev.name,
                    address=dev.address,
                    channel=comp["channel"],
                    role=comp["role"],
                )
                data["components"].append(h)
                cmps[h] = c
            pips[data["name"]] = Pipeline(**data)
    return pips, cmps


def _updater(context, port, cmd, params):
    dreq = context.socket(zmq.REQ)
    dreq.connect(f"tcp://127.0.0.1:{port}")
    dreq.send_pyobj(dict(cmd=cmd, params=params, sender=f"{__name__}._updater"))
    ret = dreq.recv_pyobj()
    dreq.close()
    return ret


def _status_helper(daemon: Daemon, yaml: bool, stgrp: str):
    if stgrp == "tomato":
        rep = Reply(
            success=True,
            msg=f"tomato running on port {daemon.port}",
            data=daemon,
        )
    elif stgrp == "pipelines":
        if yaml:
            rep = Reply(
                success=True,
                msg=f"tomato running on port {daemon.port}",
                data=daemon.pips,
            )
        else:
            ii = []
            for i in daemon.pips.values():
                line = f"name:{i.name}\tready:{i.ready}\tsampleid:{i.sampleid}\tjobid:{i.jobid}"
                ii.append(line)
            if len(ii) == 0:
                msg = f"tomato running on port {daemon.port} with no pipelines"
            else:
                msg = f"tomato running on port {daemon.port} with the following pipelines:\n\t "
                msg += "\n\t ".join(ii)
            rep = Reply(success=True, msg=msg)
    elif stgrp == "drivers":
        if yaml:
            rep = Reply(
                success=True,
                msg=f"tomato running on port {daemon.port}",
                data=daemon.drvs,
            )
        else:
            ii = []
            for i in daemon.drvs.values():
                line = f"name:{i.name}\tport:{i.port}\tpid:{i.pid}\tversion:{i.version}"
                ii.append(line)
            if len(ii) == 0:
                msg = f"tomato running on port {daemon.port} with no drivers"
            else:
                msg = f"tomato running on port {daemon.port} with the following drivers:\n\t "
                msg += "\n\t ".join(ii)
            rep = Reply(success=True, msg=msg)
    elif stgrp == "devices":
        if yaml:
            rep = Reply(
                success=True,
                msg=f"tomato running on port {daemon.port}",
                data=daemon.devs,
            )
        else:
            ii = []
            for i in daemon.devs.values():
                line = f"name:{i.name}\tdriver:{i.driver}\taddress:{i.address}\tchannels:{i.channels}"
                ii.append(line)
            if len(ii) == 0:
                msg = f"tomato running on port {daemon.port} with no devices"
            else:
                msg = f"tomato running on port {daemon.port} with the following devices:\n\t "
                msg += "\n\t ".join(ii)
            rep = Reply(success=True, msg=msg)
    elif stgrp == "components":
        if yaml:
            rep = Reply(
                success=True,
                msg=f"tomato running on port {daemon.port}",
                data=daemon.cmps,
            )
        else:
            ii = []
            for i in daemon.cmps.values():
                line = f"name:{i.name}\tdriver:{i.driver}\tdevice:{i.device}\trole:{i.role}"
                ii.append(line)
            if len(ii) == 0:
                msg = f"tomato running on port {daemon.port} with no components"
            else:
                msg = f"tomato running on port {daemon.port} with the following components:\n\t "
                msg += "\n\t ".join(ii)
            rep = Reply(success=True, msg=msg)
    return rep


def status(
    *,
    port: int,
    timeout: int,
    context: zmq.Context,
    stgrp: str = "tomato",
    yaml: bool = True,
    **_: dict,
) -> Reply:
    """
    Get status of the tomato daemon.

    Examples
    --------

    >>> # Status with a running daemon
    >>> tomato status
    Success: tomato running on port 1234

    >>> # Status without a running daemon
    >>> tomato status
    Failure: tomato not running on port 1234

    >>> # Status of a running daemon with data
    >>> tomato status -y
    data:
      appdir: /home/kraus/.config/tomato/1.0rc2.dev2
      cmps:
        [...]
      devs:
        [...]
      drvs:
        [...]
      jobs:
        [...]
      logdir: /home/kraus/.cache/tomato/1.0rc2.dev2/log
      nextjob: 1
      pips:
        [...]
      port: 1234
      settings:
        [...]
      status: running
      verbosity: 20
    msg: tomato running on port 1234
    success: true

    >>> # Status of all configured pipelines
    >>> tomato status pipelines
    Success: tomato running on port 1234 with the following pipelines:
         name:pip-counter       ready:False     sampleid:counter_1_0.1  jobid:None

    >>> # Status of all configured drivers
    >>> tomato status drivers
    Success: tomato running on port 1234 with the following drivers:
         name:example_counter   port:34747      pid:192318

    >>> # Status of all configured devices
    >>> tomato status devices
    Success: tomato running on port 1234 with the following devices:
         name:dev-counter       driver:example_counter  address:example-addr    channels:['1']

    >>> # Status of all configured components:
    >>> tomato status components
    Success: tomato running on port 1234 with the following components:
         name:example_counter:(example-addr,1)  driver:example_counter  device:dev-counter      role:counter

    """
    logger.debug("checking status of tomato on port %d", port)
    req = context.socket(zmq.REQ)
    req.connect(f"tcp://127.0.0.1:{port}")
    req.send_pyobj(dict(cmd="status", sender=f"{__name__}.status"))
    poller = zmq.Poller()
    poller.register(req, zmq.POLLIN)
    events = dict(poller.poll(timeout))
    if req in events:
        rep = req.recv_pyobj()
        return _status_helper(daemon=rep.data, yaml=yaml, stgrp=stgrp)
    else:
        req.setsockopt(zmq.LINGER, 0)
        req.close()
        return Reply(
            success=False,
            msg=f"tomato not running on port {port}",
        )


def start(
    *,
    port: int,
    timeout: int,
    context: zmq.Context,
    appdir: str,
    verbosity: int,
    **_: dict,
) -> Reply:
    """
    Start the tomato daemon.

    Examples
    --------

    >>> # Start tomato
    >>> tomato start
    Success: tomato on port 1234 reloaded with settings from /home/kraus/.config/tomato/1.0rc2.dev2

    >>> # Start tomato with a custom port
    >>> tomato start -p 1235
    Success: tomato on port 1235 reloaded with settings from /home/kraus/.config/tomato/1.0rc2.dev2

    >>> # Start tomato with another tomato running
    >>> tomato start
    Failure: required port 1234 is already in use, choose a different one

    """
    logger.debug("checking for availability of port %d", port)
    try:
        rep = context.socket(zmq.REP)
        rep.bind(f"tcp://127.0.0.1:{port}")
        stat = status(port=port, timeout=1000, context=context)
        rep.unbind(f"tcp://127.0.0.1:{port}")
        rep.setsockopt(zmq.LINGER, 0)
        rep.close()
        if stat.success:
            return Reply(
                success=False,
                msg=f"tomato-daemon already running on port {port}",
            )
    except zmq.error.ZMQError:
        return Reply(
            success=False,
            msg=f"required port {port} is already in use, choose a different one",
        )

    if not (Path(appdir) / "settings.toml").exists():
        return Reply(
            success=False,
            msg=f"settings file not found in {appdir}, run 'tomato init' to create one",
        )

    logger.debug("starting tomato on port %d", port)
    cmd = [
        "tomato-daemon",
        "-p",
        f"{port}",
        "-A",
        f"{appdir}",
        "-V",
        f"{verbosity}",
    ]
    if psutil.WINDOWS:
        cfs = subprocess.CREATE_NEW_PROCESS_GROUP
        subprocess.Popen(cmd, creationflags=cfs)
    elif psutil.POSIX:
        subprocess.Popen(cmd, start_new_session=True)
    kwargs = dict(port=port, timeout=max(timeout, 5000), context=context)
    stat = status(**kwargs)
    if stat.success:
        return reload(**kwargs, appdir=appdir)
    else:
        return Reply(
            success=False,
            msg=f"failed to start tomato on port {port}: {stat.msg}",
            data=stat.data,
        )


def stop(
    *,
    port: int,
    timeout: int,
    context: zmq.Context,
    **_: dict,
) -> Reply:
    """
    Stop a running tomato daemon.

    Will not stop the daemon if any jobs are running. Will create a state snapshot.

    Examples
    --------

    >>> # Stop tomato daemon without running jobs
    >>> tomato stop
    Success: tomato on port 1234 closed successfully

    >>> # Attempt to stop tomato daemon with running jobs
    >>> tomato stop
    Failure: jobs are running

    >>> # Attempt to stop tomato daemon which is not running
    >>> tomato stop -p 1235
    Failure: tomato not running on port 1235

    """
    stat = status(port=port, timeout=timeout, context=context)
    if stat.success:
        req = context.socket(zmq.REQ)
        req.connect(f"tcp://127.0.0.1:{port}")
        req.send_pyobj(dict(cmd="stop"))
        rep = req.recv_pyobj()
        if rep.success:
            return Reply(success=True, msg=f"tomato on port {port} closed successfully")
        else:
            return rep
    else:
        return stat


def init(
    *,
    appdir: str,
    datadir: str,
    logdir: str,
    **_: dict,
) -> Reply:
    """
    Create a default settings.toml file.

    Will overwrite any existing settings.toml file.

    Examples
    --------

    >>> tomato init
    Success: wrote default settings into /home/kraus/.config/tomato/1.0rc2.dev2/settings.toml

    """
    appdir = Path(appdir).resolve()
    datadir = Path(datadir).resolve()
    logdir = Path(logdir).resolve()
    storage = datadir.resolve() / "Jobs"
    dbpath = storage / "dbpath.sqlite"
    defaults = textwrap.dedent(
        f"""\
        # Default settings for tomato-{VERSION}
        # Generated on {str(datetime.now(timezone.utc))}
        datadir = '{datadir}'
        logdir = '{logdir}'

        [jobs]
        storage = '{storage}'
        dbpath = '{dbpath}'

        [devices]
        config = '{appdir / "devices.yml"}'

        [drivers]
        example_counter.idle_measurement_interval = 1
        """
    )
    if not appdir.exists():
        logger.debug("creating directory '%s'", appdir)
        os.makedirs(appdir)
    with (appdir / "settings.toml").open("w", encoding="utf-8") as of:
        of.write(defaults)
    if not logdir.exists():
        logger.debug("creating directory '%s'", logdir)
        os.makedirs(logdir)
    if not storage.exists():
        logger.debug("creating directory '%s'", storage)
        os.makedirs(storage)
    if not dbpath.exists():
        jobdb_setup(dbpath)
    return Reply(
        success=True,
        msg=f"wrote default settings into {appdir / 'settings.toml'}",
    )


def reload(
    *,
    port: int,
    timeout: int,
    context: zmq.Context,
    appdir: str,
    **_: dict,
) -> Reply:
    """
    Reload settings.toml and devices.yaml files and reconfigure tomato daemon.

    Examples
    --------

    >>> # Reload with compatible changes
    >>> tomato reload
    Success: tomato on port 1234 reloaded with settings from /home/kraus/.config/tomato/1.0rc2.dev2

    """
    kwargs = dict(port=port, timeout=timeout, context=context)
    logger.debug("Loading settings.toml file from %s.", appdir)
    try:
        settings = toml.load(Path(appdir) / "settings.toml")
    except FileNotFoundError:
        return Reply(
            success=False,
            msg=f"settings file not found in {appdir}, run 'tomato init' to create one",
        )

    devicefile = load_device_file(Path(settings["devices"]["config"]))
    devs = {dev["name"]: Device(**dev) for dev in devicefile["devices"]}
    pips, cmps = get_pipelines(devs, devicefile["pipelines"])
    drvs = {dev.driver: Driver(name=dev.driver) for dev in devs.values()}
    logger.debug(f"{pips=}")
    logger.debug(f"{cmps=}")
    logger.debug(f"{devs=}")
    logger.debug(f"{drvs=}")
    for drv in drvs.keys():
        if drv in settings["drivers"]:
            drvs[drv].settings.update(settings["drivers"][drv])
    ret = status(**kwargs)
    if not ret.success:
        return ret
    req = context.socket(zmq.REQ)
    req.connect(f"tcp://127.0.0.1:{port}")
    req.send_pyobj(
        dict(
            cmd="setup",
            settings=settings,
            pips=pips,
            devs=devs,
            drvs=drvs,
            cmps=cmps,
            sender=f"{__name__}.reload",
        )
    )
    ret = req.recv_pyobj()
    if ret.success:
        return Reply(
            success=True,
            msg=f"tomato on port {port} reloaded with settings from {appdir}",
            data=ret.data,
        )
    else:
        return Reply(
            success=False,
            msg=f"tomato on port {port} could not be reloaded: {ret.msg}",
            data=ret.data,
        )


def pipeline_load(
    *,
    port: int,
    timeout: int,
    context: zmq.Context,
    pipeline: str,
    sampleid: str,
    **_: dict,
) -> Reply:
    """
    Load a sample into a pipeline. Usage:

    .. code:: bash

        tomato pipeline load <pipeline> <sampleid>

    """
    stat = status(port=port, timeout=timeout, context=context)
    if not stat.success:
        return stat

    if pipeline not in stat.data.pips:
        return Reply(success=False, msg=f"pipeline {pipeline!r} not found on tomato")
    pip = stat.data.pips[pipeline]

    if pip.sampleid is not None:
        return Reply(
            success=False, msg=f"pipeline {pipeline!r} is not empty, aborting", data=pip
        )

    req = context.socket(zmq.REQ)
    req.connect(f"tcp://127.0.0.1:{port}")
    req.send_pyobj(
        dict(
            cmd="pipeline",
            params=dict(sampleid=sampleid, name=pipeline),
            sender=f"{__name__}.pipeline_load",
        ),
    )
    msg = req.recv_pyobj()
    return Reply(
        success=True, msg=f"loaded {sampleid!r} into {pipeline!r}", data=msg.data
    )


def pipeline_eject(
    *,
    port: int,
    timeout: int,
    context: zmq.Context,
    pipeline: str,
    **_: dict,
) -> Reply:
    """
    Eject any sample present in a pipeline. Usage:

    .. code:: bash

        tomato pipeline eject <pipeline>

    """
    stat = status(port=port, timeout=timeout, context=context)
    if not stat.success:
        return stat

    if pipeline not in stat.data.pips:
        return Reply(
            success=False,
            msg=f"pipeline {pipeline!r} not found on tomato",
            data=stat.data.pips,
        )
    pip = stat.data.pips[pipeline]

    if pip.sampleid is None:
        return Reply(
            success=True, msg=f"pipeline {pipeline!r} was already empty", data=pip
        )

    if pip.jobid is not None:
        return Reply(
            success=False, msg="cannot eject from a running pipeline", data=pip
        )

    req = context.socket(zmq.REQ)
    req.connect(f"tcp://127.0.0.1:{port}")
    req.send_pyobj(
        dict(
            cmd="pipeline",
            params=dict(sampleid=None, ready=False, name=pipeline),
            sender=f"{__name__}.pipeline_eject",
        )
    )
    rep = req.recv_pyobj()
    return Reply(
        success=True, msg=f"pipeline {pipeline!r} ejected succesffully", data=rep.data
    )


def pipeline_ready(
    *,
    port: int,
    timeout: int,
    context: zmq.Context,
    pipeline: str,
    **_: dict,
) -> Reply:
    """
    Mark pipeline as ready. Usage:

    .. code:: bash

        pipeline ready <pipeline>

    """
    stat = status(port=port, timeout=timeout, context=context)
    if not stat.success:
        return stat

    if pipeline not in stat.data.pips:
        return Reply(
            success=False,
            msg=f"pipeline {pipeline!r} not found on tomato",
            data=stat.data.pips,
        )
    pip = stat.data.pips[pipeline]

    if pip.ready:
        return Reply(
            success=True, msg=f"pipeline {pipeline!r} was already ready", data=pip
        )

    if pip.jobid is not None:
        return Reply(
            success=False, msg="cannot mark a running pipeline as ready", data=pip
        )

    req = context.socket(zmq.REQ)
    req.connect(f"tcp://127.0.0.1:{port}")
    req.send_pyobj(
        dict(
            cmd="pipeline",
            params=dict(ready=True, name=pipeline),
            sender=f"{__name__}.pipeline_ready",
        )
    )
    rep = req.recv_pyobj()
    return Reply(success=True, msg=f"pipeline {pipeline!r} set as ready", data=rep.data)



"""
**DriverInterface-1.0**
-----------------------
.. codeauthor::
    Peter Kraus

.. warning::

    This version of DriverInterface is deprecated and will not be supported from ``tomato-3.0`` onwards.


First prototype of the DriverInterface.

"""

from abc import ABCMeta, abstractmethod
from typing import TypeVar, Any
from pydantic import BaseModel
from threading import Thread, current_thread, RLock
from queue import Queue
from tomato.models import Reply
from dgbowl_schemas.tomato.payload import Task
import logging
from functools import wraps
from xarray import Dataset
from collections import defaultdict
import time
import atexit

logger = logging.getLogger(__name__)


def in_devmap(func):
    @wraps(func)
    def wrapper(self, **kwargs):
        if "key" in kwargs:
            key = kwargs.pop("key")
        else:
            address = kwargs.get("address")
            channel = kwargs.get("channel")
            key = (address, channel)
        if key not in self.devmap:
            msg = f"dev with address {address!r} and channel {channel} is unknown"
            return Reply(success=False, msg=msg, data=self.devmap.keys())
        return func(self, **kwargs, key=key)

    return wrapper


T = TypeVar("T")


class Attr(BaseModel):
    """A Pydantic :class:`BaseModel` used to describe device attributes."""

    type: T
    rw: bool = False
    status: bool = False
    units: str = None


class ModelInterface(metaclass=ABCMeta):
    """
    An abstract base class specifying the driver interface.

    Individual driver modules should expose a :class:`DriverInterface` which inherits
    from this abstract class. Only the methods of this class should be used to interact
    with drivers and their devices.
    """

    version: str = "1.0"

    class DeviceManager(metaclass=ABCMeta):
        """
        An abstract base class specifying a manager for an individual component.

        This class should handle determining attributes and capabilities of the component,
        the reading/writing of those attributes, processing of tasks, and caching and
        returning of task data.
        """

        driver: "ModelInterface"
        """The parent :class:`DriverInterface` instance."""

        data: dict[str, list]
        """Container for cached data on this component."""

        datalock: RLock
        """Lock object for thread-safe data manipulation."""

        key: tuple
        """The key in :obj:`self.driver.devmap` referring to this object."""

        thread: Thread
        """The worker :class:`Thread`."""

        task_list: Queue
        """A :class:`Queue` used to pass :class:`Tasks` to the worker :class:`Thread`."""

        running: bool

        def __init__(self, driver, key, **kwargs):
            self.driver = driver
            self.key = key
            self.task_list = Queue()
            self.thread = Thread(target=self.task_runner, daemon=True)
            self.data = defaultdict(list)
            self.running = False
            self.datalock = RLock()
            atexit.register(self.reset)

        def run(self):
            """Helper function for starting the :obj:`self.thread`."""
            self.thread.do_run = True
            self.thread.start()
            self.running = True

        def task_runner(self):
            """
            Target function for the :obj:`self.thread`.

            This function waits for a :class:`Task` passed using :obj:`self.task_list`,
            then handles setting all :class:`Attrs` using the :func:`prepare_task`
            function, and finally handles the main loop of the task, periodically running
            the :func:`do_task` function (using `task.sampling_interval`) until the
            maximum task duration (i.e. `task.max_duration`) is exceeded.

            The :obj:`self.thread` is re-primed for future :class:`Tasks` at the end
            of this function.
            """
            thread = current_thread()
            task: Task = self.task_list.get()
            self.prepare_task(task)
            t_start = time.perf_counter()
            t_prev = t_start
            self.data = defaultdict(list)
            while getattr(thread, "do_run"):
                t_now = time.perf_counter()
                if t_now - t_prev > task.sampling_interval:
                    with self.datalock:
                        self.do_task(task, t_start=t_start, t_now=t_now, t_prev=t_prev)
                    t_prev += task.sampling_interval
                if t_now - t_start > task.max_duration:
                    break
                time.sleep(max(1e-2, task.sampling_interval / 20))

            self.task_list.task_done()
            self.running = False
            self.thread = Thread(target=self.task_runner, daemon=True)
            logger.info(
                "task '%s' on component %s is done", task.technique_name, self.key
            )

        def prepare_task(self, task: Task, **kwargs: dict):
            """
            Given a :class:`Task`, prepare this component for execution by settin all
            :class:`Attrs` as specified in the `task.task_params` dictionary.
            """
            if task.task_params is not None:
                for k, v in task.task_params.items():
                    self.set_attr(attr=k, val=v)

        @abstractmethod
        def do_task(self, task: Task, **kwargs: dict):
            """
            Periodically called task execution function.

            This function is responsible for updating :obj:`self.data` with new data, i.e.
            performing the measurement. It should also update the values of all
            :class:`Attrs`, so that the component status is consistent with the cached data.
            """
            pass

        def stop_task(self, **kwargs: dict):
            """Stops the currently running task."""
            logger.info("stopping running task on component %s", self.key)
            setattr(self.thread, "do_run", False)

        @abstractmethod
        def set_attr(self, attr: str, val: Any, **kwargs: dict):
            """Sets the specified :class:`Attr` to `val`."""
            pass

        @abstractmethod
        def get_attr(self, attr: str, **kwargs: dict) -> Any:
            """Reads the value of the specified :class:`Attr`."""
            pass

        def get_data(self, **kwargs: dict) -> dict[str, list]:
            """Returns the cached :obj:`self.data` before clearing the cache."""
            with self.datalock:
                ret = self.data
                self.data = defaultdict(list)
            return ret

        @abstractmethod
        def attrs(**kwargs) -> dict[str, Attr]:
            """Returns a :class:`dict` of all available :class:`Attrs`."""
            pass

        @abstractmethod
        def capabilities(**kwargs) -> set:
            """Returns a :class:`set` of all supported techniques."""
            pass

        def status(self, **kwargs) -> dict:
            """Compiles a status report from :class:`Attrs` marked as `status=True`."""
            status = {}
            for attr, props in self.attrs().items():
                if props.status:
                    status[attr] = self.get_attr(attr)
            return status

        def reset(self, **kwargs) -> None:
            """Resets the component to an initial status."""
            logger.info("resetting component %s", self.key)
            self.task_list = Queue()
            self.thread = Thread(target=self.task_runner, daemon=True)
            self.data = defaultdict(list)
            self.running = False
            self.datalock = RLock()

    def CreateDeviceManager(self, key, **kwargs):
        """
        A factory function which is used to pass this instance of the :class:`ModelInterface`
        to the new :class:`DeviceManager` instance.
        """
        return self.DeviceManager(self, key, **kwargs)

    devmap: dict[tuple, DeviceManager]
    """Map of registered devices, the tuple keys are `component = (address, channel)`"""

    settings: dict[str, Any]
    """A settings map to contain driver-specific settings such as `dllpath` for BioLogic"""

    def __init__(self, settings=None):
        self.devmap = {}
        self.settings = settings if settings is not None else {}

    def dev_register(self, address: str, channel: str, **kwargs: dict) -> Reply:
        """
        Register a new device component in this driver.

        Creates a :class:`DeviceManager` representing a device component, storing it in
        the :obj:`self.devmap` using the provided `address` and `channel`.

        The returned :class:`Reply` should contain the capabilities of the registered
        component in the ``data`` slot.
        """
        key = (address, channel)
        self.devmap[key] = self.CreateDeviceManager(key, **kwargs)
        capabs = self.devmap[key].capabilities()
        return Reply(
            success=True,
            msg=f"device {key!r} registered",
            data=capabs,
        )

    @in_devmap
    def dev_teardown(self, key: tuple, **kwargs: dict) -> Reply:
        """
        Emergency stop function.

        Should set the device component into a documented, safe state. The function is
        to be only called in case of critical errors, or when the component is being
        removed, not as part of normal operation (i.e. it is not intended as a clean-up
        after task completion).
        """
        status = self.task_status(key=key, **kwargs)
        if status.data:
            logger.warning("tearing down component %s with a running task!", key)
            self.task_stop(key=key, **kwargs)
        self.dev_reset(key=key, **kwargs)
        del self.devmap[key]
        return Reply(
            success=True,
            msg=f"device {key!r} torn down",
        )

    @in_devmap
    def dev_reset(self, key: tuple, **kwargs: dict) -> Reply:
        """
        Component reset function.

        Should set the device component into a documented, safe state. This function
        is executed at the end of every job.
        """
        self.devmap[key].reset()
        return Reply(
            success=True,
            msg=f"component {key!r} reset successfully",
        )

    @in_devmap
    def dev_set_attr(self, attr: str, val: Any, key: tuple, **kwargs: dict) -> Reply:
        """
        Set value of the :class:`Attr` of the specified device component.

        Pass-through to the :func:`DeviceManager.set_attr` function. No type or
        read-write validation performed here!
        """
        self.devmap[key].set_attr(attr=attr, val=val, **kwargs)
        return Reply(
            success=True,
            msg=f"attr {attr!r} of component {key} set to {val}",
            data=val,
        )

    @in_devmap
    def dev_get_attr(self, attr: str, key: tuple, **kwargs: dict) -> Reply:
        """
        Get value of the :class:`Attr` from the specified device component.

        Pass-through to the :func:`DeviceManager.get_attr` function. Units are not
        returned; those can be queried for all :class:`Attrs` using :func:`self.attrs`.

        """
        ret = self.devmap[key].get_attr(attr=attr, **kwargs)
        return Reply(
            success=True,
            msg=f"attr {attr!r} of component {key} is: {ret}",
            data=ret,
        )

    @in_devmap
    def dev_status(self, key: tuple, **kwargs: dict) -> Reply:
        """
        Get the status report from the specified device component.

        Iterates over all :class:`Attrs` on the component that have ``status=True`` and
        returns their values in the :obj:`Reply.data` as a :class:`dict`.
        """
        ret = {}
        for k, attr in self.devmap[key].attrs(key=key, **kwargs).items():
            if attr.status:
                ret[k] = self.devmap[key].get_attr(attr=k, **kwargs)

        ret["running"] = self.devmap[key].running
        return Reply(
            success=True,
            msg=f"component {key} is{' ' if ret['running'] else ' not '}running",
            data=ret,
        )

    @in_devmap
    def task_start(self, key: tuple, task: Task, **kwargs) -> Reply:
        """
        Submit a :class:`Task` onto the specified device component.

        Pushes the supplied :class:`Task` into the :class:`Queue` of the component,
        then starts the worker thread. Checks that the :class:`Task` is among the
        capabilities of this component.
        """
        logger.info("starting task '%s' on component %s", task.technique_name, key)
        if task.technique_name not in self.devmap[key].capabilities(**kwargs):
            return Reply(
                success=False,
                msg=f"unknown task {task.technique_name!r} requested",
                data=self.capabilities(key=key),
            )

        self.devmap[key].task_list.put(task)
        self.devmap[key].run()
        logger.info("task '%s' on component %s started", task.technique_name, key)
        return Reply(
            success=True,
            msg=f"task {task!r} started successfully",
            data=task,
        )

    @in_devmap
    def task_status(self, key: tuple, **kwargs: dict) -> Reply:
        """
        Returns the task readiness status of the specified device component.

        The `running` entry in the data slot of the :class:`Reply` indicates whether
        a :class:`Task` is running. The `can_submit` entry indicates whether another
        :class:`Task` can be queued onto the device component already.
        """
        running = self.devmap[key].running
        data = dict(running=running, can_submit=not running)
        if running:
            return Reply(success=True, msg="running", data=data)
        else:
            return Reply(success=True, msg="ready", data=data)

    @in_devmap
    def task_stop(self, key: tuple, **kwargs) -> Reply:
        """
        Stops a running task and returns any collected data.

        Pass-through to :func:`DriverManager.stop_task` and :func:`task_data`.
        """
        ret = self.devmap[key].stop_task(**kwargs)
        if ret is not None:
            return Reply(success=False, msg="failed to stop task", data=ret)

        ret = self.task_data(self, key=key)
        if ret.success:
            return Reply(success=True, msg=f"task stopped, {ret.msg}", data=ret.data)
        else:
            return Reply(success=True, msg=f"task stopped, {ret.msg}")

    @in_devmap
    def task_data(self, key: tuple, **kwargs) -> Reply:
        """
        Return cached task data on the device component and clean the cache.

        Pass-through for :func:`DeviceManager.get_data`, with the caveat that the
        :class:`dict[list]` which is returned from the component is here converted to a
        :class:`Dataset` and annotated using units from :func:`attrs`.

        This function gets called by the job thread every `device.pollrate`, it therefore
        incurs some IPC cost.

        """
        data = self.devmap[key].get_data(**kwargs)

        if len(data) == 0:
            return Reply(success=False, msg="found no new datapoints")

        attrs = self.devmap[key].attrs(**kwargs)
        uts = {"uts": data.pop("uts")}
        data_vars = {}
        for k, v in data.items():
            units = {} if attrs[k].units is None else {"units": attrs[k].units}
            data_vars[k] = ("uts", v, units)
        ds = Dataset(data_vars=data_vars, coords=uts)
        return Reply(success=True, msg=f"found {len(data)} new datapoints", data=ds)

    def status(self) -> Reply:
        """
        Returns the driver status. Currently that is the names of the components in
        the `devmap`.

        """
        devkeys = self.devmap.keys()
        return Reply(
            success=True,
            msg=f"driver running with {len(devkeys)} devices",
            data=dict(devkeys=devkeys),
        )

    def reset(self) -> Reply:
        """
        Resets the driver.

        Called when the driver process is quitting. Instructs all remaining tasks to
        stop. Warns when devices linger. Passes through to :func:`dev_reset`. This is
        not a pass-through to :func:`dev_teardown`.

        """
        logger.info("resetting all components on this driver")
        for key, dev in self.devmap.items():
            if dev.thread.is_alive():
                logger.warning("stopping task on component %s", key)
                setattr(dev.thread, "do_run", False)
                dev.thread.join(timeout=1)
            if dev.thread.is_alive():
                logger.error("task on component %s is still running", key)
            else:
                logger.debug("component %s has no running task", key)
            self.dev_reset(key=key)
        return Reply(
            success=True,
            msg="all components on driver have been reset",
        )

    @in_devmap
    def capabilities(self, key: tuple, **kwargs) -> Reply:
        """
        Returns the capabilities of the device component.

        Pass-through to :func:`DriverManager.capabilities`.
        """
        ret = self.devmap[key].capabilities(**kwargs)
        return Reply(
            success=True,
            msg=f"capabilities supported by component {key!r} are: {ret}",
            data=ret,
        )

    @in_devmap
    def attrs(self, key: tuple, **kwargs: dict) -> Reply:
        """
        Query available :class:`Attrs` on the specified device component.

        Pass-through to the :func:`DeviceManager.attrs` function.
        """
        ret = self.devmap[key].attrs(**kwargs)
        return Reply(
            success=True,
            msg=f"attrs of component {key!r} are: {ret}",
            data=ret,
        )



import os
import pytest
import zmq

from tomato import ketchup, tomato
from . import utils

PORT = 12345
CTXT = zmq.Context()

kwargs = dict(port=PORT, timeout=1000, context=CTXT)


@pytest.mark.parametrize(
    "pl, jn",
    [
        ("counter_1_0.1.yml", None),
        ("counter_1_0.1.yml", "counter"),
    ],
)
def test_ketchup_submit_one(pl, jn, datadir, start_tomato_daemon, stop_tomato_daemon):
    os.chdir(datadir)
    assert utils.wait_until_tomato_running(port=PORT, timeout=1000)

    ret = ketchup.submit(**kwargs, payload=pl, jobname=jn)
    utils.wait_until_ketchup_status(jobid=1, status="q", port=PORT, timeout=1000)

    print(f"{ret=}")
    assert ret.success
    assert ret.data.id == 1
    assert ret.data.jobname == jn


def test_ketchup_submit_two(datadir, start_tomato_daemon, stop_tomato_daemon):
    os.chdir(datadir)
    assert utils.wait_until_tomato_running(port=PORT, timeout=1000)

    ret = ketchup.submit(payload="counter_1_0.1.yml", jobname="job-1", **kwargs)
    ret = ketchup.submit(payload="counter_5_0.2.yml", jobname="job-2", **kwargs)
    utils.wait_until_ketchup_status(jobid=2, status="q", port=PORT, timeout=1000)

    print(f"{ret=}")
    assert ret.success
    assert ret.data.id == 2
    assert ret.data.jobname == "job-2"


def test_ketchup_status_empty(start_tomato_daemon, stop_tomato_daemon):
    assert utils.wait_until_tomato_running(port=PORT, timeout=5000)
    ret = ketchup.status(**kwargs, verbosity=0, jobids=[])
    print(f"{ret=}")
    assert not ret.success
    assert "job queue is empty" in ret.msg


def test_ketchup_status_all_queued(datadir, start_tomato_daemon, stop_tomato_daemon):
    os.chdir(datadir)
    assert utils.wait_until_tomato_running(port=PORT, timeout=1000)

    ret = ketchup.submit(payload="counter_1_0.1.yml", jobname="job-1", **kwargs)
    ret = ketchup.submit(payload="counter_5_0.2.yml", jobname="job-2", **kwargs)
    utils.wait_until_ketchup_status(jobid=2, status="q", port=PORT, timeout=1000)

    ret = ketchup.status(**kwargs, verbosity=0, jobids=[])
    print(f"{ret=}")
    assert ret.success
    assert "found 2" in ret.msg
    assert len(ret.data) == 2


def test_ketchup_status_one_queued(datadir, start_tomato_daemon, stop_tomato_daemon):
    os.chdir(datadir)
    assert utils.wait_until_tomato_running(port=PORT, timeout=1000)

    ret = ketchup.submit(payload="counter_1_0.1.yml", jobname="job-1", **kwargs)
    ret = ketchup.submit(payload="counter_5_0.2.yml", jobname="job-2", **kwargs)
    utils.wait_until_ketchup_status(jobid=2, status="q", port=PORT, timeout=1000)

    ret = ketchup.status(**kwargs, verbosity=0, jobids=[2])
    print(f"{ret=}")
    assert ret.success
    assert "found 1" in ret.msg
    assert len(ret.data) == 1
    assert ret.data[0].id == 2


def test_ketchup_status_two_queued(datadir, start_tomato_daemon, stop_tomato_daemon):
    os.chdir(datadir)
    assert utils.wait_until_tomato_running(port=PORT, timeout=1000)

    ret = ketchup.submit(payload="counter_1_0.1.yml", jobname="job-1", **kwargs)
    ret = ketchup.submit(payload="counter_5_0.2.yml", jobname="job-2", **kwargs)
    utils.wait_until_ketchup_status(jobid=2, status="q", port=PORT, timeout=1000)

    ret = ketchup.status(**kwargs, verbosity=0, jobids=[1, 2])
    print(f"{ret=}")
    assert ret.success
    assert "found 2" in ret.msg
    assert len(ret.data) == 2
    assert ret.data[0].id == 1
    assert ret.data[1].id == 2


def test_ketchup_status_running(datadir, start_tomato_daemon, stop_tomato_daemon):
    os.chdir(datadir)
    utils.run_casenames(["counter_5_0.2"], [None], ["pip-counter"])
    assert utils.wait_until_ketchup_status(jobid=1, status="r", port=PORT, timeout=5000)

    ret = ketchup.status(**kwargs, verbosity=0, jobids=[1])
    print(f"{ret=}")
    assert ret.success
    assert "found 1" in ret.msg
    assert len(ret.data) == 1
    assert ret.data[0].status == "r"


def test_ketchup_status_complete(datadir, start_tomato_daemon, stop_tomato_daemon):
    os.chdir(datadir)
    utils.run_casenames(["counter_1_0.1"], [None], ["pip-counter"])
    utils.wait_until_ketchup_status(jobid=1, status="r", port=PORT, timeout=5000)

    assert utils.wait_until_ketchup_status(jobid=1, status="c", port=PORT, timeout=5000)
    ret = ketchup.status(**kwargs, verbosity=0, jobids=[1])
    print(f"{ret=}")
    assert ret.success
    assert ret.data[0].status == "c"
    assert os.path.exists("results.1.nc")


def test_ketchup_cancel_running(datadir, start_tomato_daemon, stop_tomato_daemon):
    os.chdir(datadir)
    utils.run_casenames(["counter_60_0.1"], [None], ["pip-counter"])
    assert utils.wait_until_ketchup_status(jobid=1, status="r", port=PORT, timeout=5000)

    assert utils.wait_until_pickle(jobid=1, timeout=2000)
    ret = ketchup.cancel(**kwargs, verbosity=0, jobids=[1])
    print(f"{ret=}")
    assert ret.success
    assert ret.data[0].status == "rd"

    assert utils.wait_until_ketchup_status(
        jobid=1, status="cd", port=PORT, timeout=5000
    )
    ret = ketchup.status(**kwargs, verbosity=0, jobids=[1])
    print(f"{ret=}")
    print(f"{os.listdir()=}")
    print(f"{os.listdir('Jobs')=}")
    print(f"{os.listdir(os.path.join('Jobs', '1'))=}")
    assert ret.data[0].status == "cd"
    assert os.path.exists("results.1.nc")


def test_ketchup_cancel_queued(datadir, start_tomato_daemon, stop_tomato_daemon):
    os.chdir(datadir)
    assert utils.wait_until_tomato_running(port=PORT, timeout=1000)
    utils.wait_until_tomato_drivers(port=PORT, timeout=3000)

    ret = ketchup.submit(payload="counter_60_0.1.yml", jobname=None, **kwargs)
    tomato.pipeline_load(**kwargs, pipeline="pip-counter", sampleid="counter_60_0.1")
    assert utils.wait_until_ketchup_status(
        jobid=1, status="qw", port=PORT, timeout=5000
    )

    ret = ketchup.cancel(**kwargs, verbosity=0, jobids=[1])
    print(f"{ret=}")
    assert ret.success
    assert ret.data[0].status == "cd"


def test_ketchup_snapshot(datadir, start_tomato_daemon, stop_tomato_daemon):
    os.chdir(datadir)
    utils.run_casenames(["counter_60_0.1"], [None], ["pip-counter"])
    assert utils.wait_until_ketchup_status(jobid=1, status="r", port=PORT, timeout=5000)

    assert utils.wait_until_pickle(jobid=1, timeout=2000)
    ret = ketchup.snapshot(jobids=[1], port=PORT, context=CTXT)
    print(f"{ret=}")
    assert ret.success
    assert os.path.exists("snapshot.1.nc")


def test_ketchup_search(datadir, start_tomato_daemon, stop_tomato_daemon):
    os.chdir(datadir)
    assert utils.wait_until_tomato_running(port=PORT, timeout=1000)

    ret = ketchup.submit(payload="counter_1_0.1.yml", jobname="job-1", **kwargs)
    ret = ketchup.submit(payload="counter_5_0.2.yml", jobname="job-2", **kwargs)

    ret = ketchup.search(jobname="2", port=PORT, context=CTXT)
    print(f"{ret=}")
    assert ret.success
    assert len(ret.data) == 1

    ret = ketchup.search(jobname="job", port=PORT, context=CTXT)
    print(f"{ret=}")
    assert ret.success
    assert len(ret.data) == 2

    ret = ketchup.search(jobname="wrong", port=PORT, context=CTXT)
    print(f"{ret=}")
    assert ret.success is False


@pytest.mark.parametrize(
    "pl, jn",
    [
        ("counter_invalid_1.yml", None),
        ("counter_invalid_2.yml", None),
        ("counter_invalid_3.yml", None),
        ("counter_invalid_4.yml", None),
    ],
)
def test_ketchup_validation(pl, jn, datadir, start_tomato_daemon, stop_tomato_daemon):
    os.chdir(datadir)
    assert utils.wait_until_tomato_running(port=PORT, timeout=1000)
    utils.wait_until_tomato_drivers(port=PORT, timeout=3000)

    ret = ketchup.submit(**kwargs, payload=pl, jobname=jn)
    print(f"{ret=}")
    assert ret.success
    assert ret.data.id == 1
    with pytest.raises(AssertionError):
        assert utils.wait_until_ketchup_status(
            jobid=1, status="qw", port=PORT, timeout=1000
        )



"""
**tomato.ketchup**: command line interface to the tomato job queue
------------------------------------------------------------------
.. codeauthor::
    Peter Kraus

Module of functions to manage the job queue of :mod:`tomato`. Includes job management
functions:

- :func:`submit` to submit a *job* to *queue*
- :func:`status` to query the status of tomato's *queue*, or a *job*
- :func:`cancel` to cancel a queued or kill a running *job*
- :func:`snapshot` to create an up-to-date FAIR data archive of a running *job*
- :func:`search` to find a ``jobid`` of a *job* from ``jobname``

.. warning::

    This module should interact with the job sqlite database only via the ``tomato.daemon.cmd``
    interface functions ``set_job()`` and ``get_jobs()``, **not via the ``jobdb`` module**.
    This is necessary to ensure users that don't have write access to the job database can
    still submit/manage their jobs.

"""

import json
import logging
from pathlib import Path
from datetime import datetime, timezone
import yaml
import zmq
from packaging.version import Version
from dgbowl_schemas.tomato import to_payload

from tomato.daemon.io import merge_netcdfs
from tomato.models import Reply

log = logging.getLogger(__name__)

__latest_payload__ = "2.0"


def submit(
    *,
    port: int,
    context: zmq.Context,
    payload: str,
    jobname: str,
    **_: dict,
) -> Reply:
    """
    Job submission function.

    Attempts to open the ``yaml/json`` file specified in the ``payload`` argument,
    and submit it to tomato's queue.

    Reply contains information about the submitted job.

    Examples
    --------

    >>> # Submit a job:
    >>> ketchup submit counter_15_0.1.yml
    Success: job submitted successfully with jobid 1

    >>> # Submit a job with a job name:
    >>> ketchup submit counter_15_0.1.yml -j jobname_is_this
    Success: job submitted successfully with jobid 1 and jobname 'jobname_is_this'

    >>> # Submit a job with yaml output:
    >>> ketchup submit counter_15_0.1.yml -y
    data:
        completed_at: null
        executed_at: null
        id: 1
        [...]
        status: q
        submitted_at: '2024-11-17 19:39:16.972593+00:00'
    msg: job submitted successfully with jobid 1
    success: true

    """
    payload = Path(payload)
    if payload.exists() and payload.is_file():
        pass
    else:
        return Reply(success=False, msg=f"payload file {payload} not found")

    with payload.open() as inf:
        if payload.suffix == ".json":
            pldict = json.load(inf)
        elif payload.suffix in {".yml", ".yaml"}:
            pldict = yaml.full_load(inf)
        else:
            return Reply(success=False, msg="payload must be a yaml or a json file")

    payload = to_payload(**pldict)
    maxver = Version(__latest_payload__)
    while hasattr(payload, "update"):
        temp = payload.update()
        if hasattr(temp, "version"):
            if Version(temp.version) > maxver:
                break
        payload = temp

    if payload.settings.output.path is None:
        cwd = str(Path().resolve())
        log.info(f"Output path not set. Setting output path to {cwd}")
        payload.settings.output.path = cwd
    if payload.settings.snapshot is not None and payload.settings.snapshot.path is None:
        cwd = str(Path().resolve())
        log.info(f"Snapshot path not set. Setting output path to {cwd}")
        payload.settings.snapshot.path = cwd

    log.debug("queueing 'payload' into 'queue'")
    req = context.socket(zmq.REQ)
    req.connect(f"tcp://127.0.0.1:{port}")
    dt = str(datetime.now(timezone.utc))
    params = dict(payload=payload, jobname=jobname, submitted_at=dt)
    req.send_pyobj(dict(cmd="set_job", id=None, params=params))
    ret = req.recv_pyobj()
    req.close()
    if ret.success:
        msg = f"job submitted successfully with jobid {ret.data.id}"
        if ret.data.jobname is not None:
            msg += f" and jobname {ret.data.jobname!r}"
        return Reply(success=True, msg=msg, data=ret.data)
    else:
        return Reply(success=False, msg="unknown error", data=ret.data)


def status(
    *,
    port: int,
    context: zmq.Context,
    verbosity: int,
    jobids: list[int],
    **_: dict,
) -> Reply:
    """
    Job status query function.

    Reply contains information about all matched jobs.

    .. note::

        Calling ``ketchup status`` with a single ``jobid`` will return a ``yaml``
        :class:`list`, even though status of only one element was queried.

    Examples
    --------

    >>> # Get status of a given job
    >>> ketchup status 1
    Success: found 1 job with status 'qw': [1]

    >>> # Get status of multiple jobs
    >>> ketchup status 1 2
    Success: found 2 job with status 'qw': [1]
             found 1 job with status 'c' : [2]

    >>> # Get status of non-existent job
    >>> ketchup status 3
    Failure: found no jobs with jobids [3]

    >>> # Get a status of a job with yaml output
    >>> ketchup status 1 -y
    data:
      - completed_at: null
        executed_at: null
        id: 1
        [...]
        status: qw
        submitted_at: '2024-11-17 17:53:46.133355+00:00'
    msg: found 1 job with status ['qw']
    success: true

    """
    req = context.socket(zmq.REQ)
    req.connect(f"tcp://127.0.0.1:{port}")

    if len(jobids) == 0:
        req.send_pyobj(dict(cmd="get_jobs", where="id IS NOT NULL"))
        rets = req.recv_pyobj().data
        if len(rets) == 0:
            return Reply(success=False, msg="job queue is empty")
    else:
        where = f"id IN ({', '.join([str(j) for j in jobids])})"
        req.send_pyobj(dict(cmd="get_jobs", where=where))
        rets = req.recv_pyobj().data
        if len(rets) == 0:
            if len(jobids) == 1:
                msg = f"found no job with jobid {jobids}"
            else:
                msg = f"found no jobs with jobids {jobids}"
            return Reply(success=False, msg=msg)

    req.close()

    if len(rets) == 1:
        msg = f"found {len(rets)} job with status {[job.status for job in rets]}"
    else:
        msg = ""
        for st in ["q", "qw", "r", "rd", "c", "cd", "ce"]:
            jobst = [j.id for j in rets if j.status == st]
            if len(jobst) > 1:
                msg += (
                    f"found {len(jobst)} jobs with status {st!r:4s}: {jobst}\n         "
                )
            elif len(jobst) == 1:
                msg += (
                    f"found {len(jobst)} job with status {st!r:4s}: {jobst}\n         "
                )
        msg = msg.strip()
    return Reply(success=True, msg=msg, data=rets)


def cancel(
    *,
    port: int,
    context: zmq.Context,
    jobids: list[int],
    **_: dict,
) -> Reply:
    """
    Job cancellation function.

    .. note::

        The :func:`~ketchup.functions.cancel` only sets the status of the running
        job to ``rd``; the actual job cancellation is performed in the
        :mod:`tomato.daemon.job` module.

    Examples
    --------

    >>> # Cancel a job:
    >>> ketchup cancel 1
    Success: job [1] cancelled successfully

    >>> # Cancel a job with yaml output:
    >>> ketchup cancel 2 -y
    data:
      - completed_at: null
        executed_at: null
        id: 2
        [...]
        status: cd
        submitted_at: '2024-03-03 15:23:50.702504+00:00'
    msg: job [2] cancelled successfully
    success: true

    """
    req = context.socket(zmq.REQ)
    req.connect(f"tcp://127.0.0.1:{port}")
    where = f"id IN ({', '.join([str(j) for j in jobids])})"
    req.send_pyobj(dict(cmd="get_jobs", where=where))
    jobs = {i.id: i for i in req.recv_pyobj().data}

    for jobid in jobids:
        if jobid not in jobs:
            return Reply(success=False, msg=f"job with jobid {jobid} does not exist")

    data = []
    for jobid in jobids:
        if jobs[jobid].status in {"q", "qw"}:
            params = dict(status="cd")
        elif jobs[jobid].status in {"r"}:
            params = dict(status="rd")
        elif jobs[jobid].status in {"cd", "ce", "c"}:
            continue
        req.send_pyobj(dict(cmd="set_job", id=jobid, params=params))
        ret = req.recv_pyobj()
        if ret.success:
            data.append(ret.data)
        else:
            return Reply(success=False, msg="unknown error", data=ret.data)

    req.close()

    if len(data) == 1:
        msg = f"job {[j.id for j in data]} cancelled successfully"
    else:
        msg = f"jobs {[j.id for j in data]} cancelled successfully"
    return Reply(success=True, msg=msg, data=data)


def snapshot(
    *,
    port: int,
    jobids: list[int],
    context: zmq.Context,
    **_: dict,
) -> Reply:
    """
    Create a snapshot of job data.

    Requests an up-to-date snapshot of the data of the job identified by ``jobid``.
    Checks whether the job is running, raises a warning if job has been finished.

    Examples
    --------

    >>> # Create a snapshot in current working directory:
    >>> ketchup snapshot 3
    Success: snapshot for job [3] created successfully

    """
    req = context.socket(zmq.REQ)
    req.connect(f"tcp://127.0.0.1:{port}")
    where = f"id IN ({', '.join([str(j) for j in jobids])})"
    req.send_pyobj(dict(cmd="get_jobs", where=where))
    jobs = {i.id: i for i in req.recv_pyobj().data}
    req.close()

    for jobid in jobids:
        if jobid not in jobs:
            return Reply(success=False, msg=f"job {jobid} does not exist")
        if jobs[jobid].status in {"q", "qw"}:
            return Reply(success=False, msg=f"job {jobid} is still queued")

    for jobid in jobids:
        jobs[jobid].snappath = f"snapshot.{jobid}.nc"
        merge_netcdfs(jobs[jobid], snapshot=True)
    if len(jobids) > 1:
        msg = f"snapshot for jobs {jobids} created successfully"
    else:
        msg = f"snapshot for job {jobids} created successfully"
    return Reply(success=True, msg=msg)


def search(
    *,
    port: int,
    jobname: str,
    context: zmq.Context,
    **_: dict,
) -> Reply:
    """
    Search the queue for a job that matches a given jobname.

    Searches the queue for a job that matches the ``jobname``, returns the
    job status and ``jobid``.

    Examples
    --------

    >>> # Search for a valid jobname
    >>> ketchup search counter
    Success: job matching 'counter' found: [1]

    >>> # Search for an invalid jobname
    >>> ketchup search nothing
    Failure: no job matching 'nothing' found

    """
    req = context.socket(zmq.REQ)
    req.connect(f"tcp://127.0.0.1:{port}")
    where = f"jobname LIKE '%{jobname}%'"
    req.send_pyobj(dict(cmd="get_jobs", where=where))
    jobs = req.recv_pyobj().data
    req.close()

    if len(jobs) > 0:
        if len(jobs) == 1:
            msg = f"job matching {jobname!r} found: {[j.id for j in jobs]}"
        else:
            msg = f"jobs matching {jobname!r} found: {[j.id for j in jobs]}"
        return Reply(success=True, msg=msg, data=jobs)
    else:
        return Reply(success=False, msg=f"no job matching {jobname!r} found")



"""
**tomato.models**: Pydantic models for internal tomato use
----------------------------------------------------------
.. codeauthor::
    Peter Kraus
"""

from pydantic import BaseModel, Field, field_validator
from typing import Optional, Any, Mapping, Sequence, Literal
from dgbowl_schemas.tomato import to_payload
from dgbowl_schemas.tomato.payload import Payload
import logging
import pickle


logger = logging.getLogger(__name__)


class Driver(BaseModel):
    name: str
    version: Optional[str] = None
    port: Optional[int] = None
    pid: Optional[int] = None
    spawned_at: Optional[str] = None
    connected_at: Optional[str] = None
    settings: Mapping[str, Any] = Field(default_factory=dict)


class Device(BaseModel):
    name: str
    driver: str
    address: str
    channels: Sequence[str]
    pollrate: int = 1

    @field_validator("channels", mode="before")
    def coerce_channels(cls, v):
        if any([isinstance(vv, int) for vv in v]):
            logger.warning(
                "Supplying 'channels' as a Sequence[int] is deprecated "
                "and will stop working in tomato-2.0."
            )
            return [str(vv) for vv in v]
        return v


class Component(BaseModel):
    name: str
    driver: str
    device: str
    address: str
    channel: str
    role: str
    capabilities: Optional[set[str]] = None

    @field_validator("channel", mode="before")
    def coerce_channel(cls, v):
        if isinstance(v, int):
            logger.warning(
                "Supplying 'channel' as an int is deprecated "
                "and will stop working in tomato-2.0."
            )
            return str(v)
        return v


class Pipeline(BaseModel):
    name: str
    ready: bool = False
    jobid: Optional[int] = None
    sampleid: Optional[str] = None
    components: Sequence[str] = Field(default_factory=list)


class Job(BaseModel):
    id: Optional[int] = None
    payload: Payload
    jobname: Optional[str] = None
    pid: Optional[int] = None
    status: Literal["q", "qw", "r", "rd", "c", "cd", "ce"] = "q"
    submitted_at: Optional[str] = None
    executed_at: Optional[str] = None
    completed_at: Optional[str] = None
    jobpath: Optional[str] = None
    respath: Optional[str] = None
    snappath: Optional[str] = None

    @field_validator("payload", mode="before")
    def coerce_payload(cls, v):
        if isinstance(v, bytes):
            v = pickle.loads(v)
        if isinstance(v, dict):
            v = to_payload(**v)
        return v


class Daemon(BaseModel, arbitrary_types_allowed=True):
    status: Literal["bootstrap", "running", "stop"]
    port: int
    verbosity: int
    appdir: str
    settings: dict
    pips: Mapping[str, Pipeline] = Field(default_factory=dict)
    devs: Mapping[str, Device] = Field(default_factory=dict)
    drvs: Mapping[str, Driver] = Field(default_factory=dict)
    cmps: Mapping[str, Component] = Field(default_factory=dict)


class Reply(BaseModel):
    success: bool
    msg: Optional[str] = None
    data: Optional[Any] = None



import os
from pathlib import Path
import zmq
import time
import psutil
import logging

from tomato import ketchup, tomato
from . import utils

PORT = 12345
CTXT = zmq.Context()
WAIT = 5000

kwargs = dict(port=PORT, timeout=1000, context=CTXT)


def test_stop_with_queued_jobs(datadir, start_tomato_daemon, stop_tomato_daemon):
    assert utils.wait_until_tomato_running(port=PORT, timeout=WAIT)
    os.chdir(datadir)
    ketchup.submit(payload="counter_1_0.1.yml", jobname="job-1", **kwargs)
    ketchup.submit(payload="counter_5_0.2.yml", jobname="job-2", **kwargs)
    time.sleep(1)
    tomato.stop(**kwargs)
    assert utils.wait_until_tomato_stopped(port=PORT, timeout=5000)
    assert os.path.exists("tomato_state_12345.pkl")

    tomato.start(**kwargs, appdir=Path(), verbosity=0)
    assert utils.wait_until_tomato_running(port=PORT, timeout=WAIT)
    ret = ketchup.status(**kwargs, jobids=[], verbosity=logging.DEBUG)
    print(f"{ret=}")
    assert ret.success
    assert len(ret.data) == 2


def test_stop_with_running_jobs(datadir, start_tomato_daemon, stop_tomato_daemon):
    assert utils.wait_until_tomato_running(port=PORT, timeout=WAIT)
    os.chdir(datadir)
    ketchup.submit(payload="counter_5_0.2.yml", jobname="job-1", **kwargs)
    tomato.pipeline_load(**kwargs, pipeline="pip-counter", sampleid="counter_5_0.2")
    tomato.pipeline_ready(**kwargs, pipeline="pip-counter")
    utils.wait_until_ketchup_status(jobid=1, status="r", port=PORT, timeout=WAIT)
    ret = tomato.stop(**kwargs)
    print(f"{ret=}")
    assert ret.success is False
    assert "jobs are running" in ret.msg


def test_restart_with_running_jobs(datadir, start_tomato_daemon, stop_tomato_daemon):
    assert utils.wait_until_tomato_running(port=PORT, timeout=WAIT)
    os.chdir(datadir)
    ketchup.submit(payload="counter_20_5.yml", jobname="job-1", **kwargs)
    tomato.pipeline_load(**kwargs, pipeline="pip-counter", sampleid="counter_20_5")
    tomato.pipeline_ready(**kwargs, pipeline="pip-counter")
    assert utils.wait_until_ketchup_status(jobid=1, status="r", port=PORT, timeout=WAIT)

    ret = tomato.stop(**kwargs)
    utils.kill_tomato_daemon(port=PORT)

    assert os.path.exists("tomato_state_12345.pkl")
    ret = tomato.start(**kwargs, appdir=Path(), verbosity=0)
    print(f"{ret=}")
    assert utils.wait_until_tomato_running(port=PORT, timeout=WAIT)
    ret = ketchup.status(**kwargs, jobids=[], verbosity=logging.DEBUG)
    print(f"{ret=}")
    assert ret.success
    assert len(ret.data) == 1
    assert ret.data[0].status == "r"

    assert utils.wait_until_ketchup_status(
        jobid=1, status="c", port=PORT, timeout=25000
    )
    ret = ketchup.status(**kwargs, jobids=[1], verbosity=logging.DEBUG)
    print(f"{ret=}")
    assert ret.success
    assert len(ret.data) == 1
    assert ret.data[0].status == "c"


def test_restart_with_complete_jobs(datadir, start_tomato_daemon, stop_tomato_daemon):
    assert utils.wait_until_tomato_running(port=PORT, timeout=WAIT)
    os.chdir(datadir)
    ketchup.submit(payload="counter_5_0.2.yml", jobname="job-1", **kwargs)
    tomato.pipeline_load(**kwargs, pipeline="pip-counter", sampleid="counter_5_0.2")
    tomato.pipeline_ready(**kwargs, pipeline="pip-counter")
    assert utils.wait_until_ketchup_status(jobid=1, status="r", port=PORT, timeout=WAIT)

    ret = tomato.stop(**kwargs)
    utils.kill_tomato_daemon(port=PORT)

    time.sleep(5)
    tomato.start(**kwargs, appdir=Path(), verbosity=0)
    assert utils.wait_until_tomato_running(port=PORT, timeout=WAIT)
    assert utils.wait_until_ketchup_status(jobid=1, status="c", port=PORT, timeout=5000)
    ret = ketchup.status(**kwargs, jobids=[1], verbosity=logging.DEBUG)
    print(f"{ret=}")
    assert ret.success
    assert len(ret.data) == 1
    assert ret.data[0].status == "c"

    ret = tomato.status(**kwargs)
    print(f"{ret=}")
    assert ret.success
    assert ret.data.pips["pip-counter"].jobid is None
    assert ret.data.pips["pip-counter"].sampleid == "counter_5_0.2"


def test_restart_with_crashed_jobs(datadir, start_tomato_daemon, stop_tomato_daemon):
    assert utils.wait_until_tomato_running(port=PORT, timeout=WAIT)
    os.chdir(datadir)
    ketchup.submit(payload="counter_20_5.yml", jobname="job-1", **kwargs)
    tomato.pipeline_load(**kwargs, pipeline="pip-counter", sampleid="counter_20_5")
    tomato.pipeline_ready(**kwargs, pipeline="pip-counter")
    utils.wait_until_ketchup_status(jobid=1, status="r", port=PORT, timeout=WAIT)
    ret = ketchup.status(**kwargs, jobids=[1], verbosity=logging.DEBUG)
    print(f"{ret=}")
    pid = ret.data[0].pid

    ret = tomato.stop(**kwargs)
    utils.kill_tomato_daemon(port=PORT)

    proc = psutil.Process(pid=pid)
    proc.terminate()
    psutil.wait_procs([proc], timeout=3)

    tomato.start(**kwargs, appdir=Path(), verbosity=0)
    assert utils.wait_until_tomato_running(port=PORT, timeout=WAIT)
    ret = ketchup.status(**kwargs, jobids=[1], verbosity=logging.DEBUG)
    print(f"{ret=}")
    assert ret.success
    assert len(ret.data) == 1
    assert ret.data[0].status == "ce"

    ret = tomato.status(**kwargs)
    print(f"{ret=}")
    assert ret.success
    assert ret.data.pips["pip-counter"].jobid is None
    assert ret.data.pips["pip-counter"].sampleid == "counter_20_5"


def test_crashed_driver_restarts(datadir, start_tomato_daemon, stop_tomato_daemon):
    assert utils.wait_until_tomato_running(port=PORT, timeout=WAIT)
    assert utils.wait_until_tomato_drivers(port=PORT, timeout=3000)
    os.chdir(datadir)

    ret = tomato.status(**kwargs, stgrp="drivers")
    assert ret.success
    print(f"{ret.data=}")
    pid = ret.data["example_counter"].pid
    p = psutil.Process(pid)
    p.terminate()
    gone, alive = psutil.wait_procs([p], timeout=5)
    print(f"{gone=}")
    print(f"{alive=}")
    time.sleep(5)

    ret = tomato.status(port=PORT, timeout=1000, context=CTXT, stgrp="drivers")
    assert ret.success
    print(f"{ret.data=}")
    assert pid != ret.data["example_counter"].pid


def test_crashed_driver_with_running_jobs(
    datadir, start_tomato_daemon, stop_tomato_daemon
):
    assert utils.wait_until_tomato_running(port=PORT, timeout=WAIT)
    os.chdir(datadir)
    ketchup.submit(payload="counter_5_0.2.yml", jobname="job-1", **kwargs)
    tomato.pipeline_load(**kwargs, pipeline="pip-counter", sampleid="counter_5_0.2")
    tomato.pipeline_ready(**kwargs, pipeline="pip-counter")
    assert utils.wait_until_ketchup_status(jobid=1, status="r", port=PORT, timeout=WAIT)

    ret = tomato.status(**kwargs, stgrp="drivers")
    assert ret.success
    print(f"{ret.data=}")
    pid = ret.data["example_counter"].pid
    p = psutil.Process(pid)
    p.terminate()
    gone, alive = psutil.wait_procs([p], timeout=5)
    print(f"{gone=}")
    print(f"{alive=}")

    utils.wait_until_ketchup_status(jobid=1, status="ce", port=PORT, timeout=5000)
    ret = ketchup.status(**kwargs, jobids=[1], verbosity=logging.DEBUG)
    print(f"{ret=}")
    assert ret.success
    assert len(ret.data) == 1
    assert ret.data[0].status == "ce"
